/* Copyright (c) 2001, Craig Cleaveland,  All rights reserved.
	http://craigc.com
	
Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the following
  disclaimer.

  Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

  Neither Craig Cleaveland nor the names of its contributors may be
  used to endorse or promote products derived from this software
  without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

// To do:
// improved error messages and error test suite
// worth improving code by having a lookahead method?

import java.io.*;
import java.util.*;

/**
 * This class translates the simple form of TL to XML
 *
 * The main method is typically invoked on the command line
 * with two arguments.  The first is the name of the input
 * file (simple form) and the second is the name of the output
 * file (XML).  For example, to translate hello.template
 * to hello.xml, you would say:<pre><code>
 *
 *      java Simple2XML hello.template hello.xml
 *</code></pre>
 * @author Craig Cleaveland
 * @version 1.0
 */
public class Simple2XML {
    static String delimiter = "#";
    static PrintWriter out;
    static InputStream in;
    static int nextchar = -2;
    static int linecount = 1;
    static int errorcount = 0;

    static String translate() {
        while (more()) {
            char c = next();
            if (c=='&') {
                out.print("&amp;");
            } else if (c=='<') {
                out.print("&lt;");
            } else if (c=='>') {
                out.print("&gt;");
            } else if (c=='#') {
                if (!more()) {
                    error("Unknown #");
                    return "";
                }
                char c2 = next();
                String reverse = "";
                if (c2=='.') {
                    reverse = " whitespace=\"reverse\"";
                    if (!more()) {
                        error("Unknown #.");
                        return "";
                    }
                    c2 = next();
                }
                
                if (c2=='#') {
                    if (reverse.equals("")) {
                        out.print("#");
                    } else {
                        out.print("<empty "+reverse+"/>");
                    }
                } else if (c2=='"') {
                    grabExpr(reverse);
                } else if (c2=='(') {
                    grabJexpr(reverse);
                } else if (c2==' ') {
                    grabJstmt(reverse);
                } else if (c2=='f') {
                    c2 = next();
                    if (c2=='o') {
                        grabFor(reverse);
                    } else if (c2=='i' && next()=='#') {
                        return "fi";
                    } else {
                        error("Unknown #f"+c2);
                    }
                } else if (c2=='i') {
                    if (next()=='f') {
                        grabIf(reverse);
                    } else {
                        error("Unknown #i");
                    }
                } else if (c2=='e') {
                    char c3 = next();
                    if (c3=='n' && next()=='d' && next()=='#') {
                        return "end";
                    } else if (c3=='l' && next()=='s' && next()=='e' && next()=='#') {
                        return "else";
                    } else {
                        error("Unknown #e"+c3);
                    }
                } else if (c2=='d') {
                    if (next()=='e' && next()=='c' && next()=='l' && next()=='a' && next()=='r') {
                        char c3 = next();
                        if (c3=='e') {
                            grabDeclare(reverse);
                        } else if (c3=='a' && next()=='t' && next()=='i' && next()=='o' && next()=='n' && next()=='s') {
                            grabDeclarations(reverse);
                        } else {
                            error("Unknown #declar...");
                        }
                    } else {
                        error("Unknown #d...");
                    }
                } else {
                    error("Unknown #"+c2);
                    return "";
                }
            } else {
                out.print(c);
            }
        }
        return "";
    }
    
    static void grabExpr(String reverse) {
        String path = getPath();
        if (path==null) {
            error("missing path");
        } else {
            out.print("<value path="+path+reverse+"/>");
        }
    }
    
    static String getPath() {
        String path = "";
        char previous = ' ';
        while (more()) {
            char c = next();
            if (c=='#' && previous=='"') {
                return "\""+path;
            }
            if (c=='#' && previous=='#') {
                previous = ' ';
                continue; // ## is translated to #
            }
            previous = c;
            path += translateChar(c);
        }
        return null;
    }
    
    static void grabJexpr(String reverse) {
        int lc = linecount;
        String expr = "";
        char previous = ' ';
        while (more()) {
            char c = next();
            if (c=='#' && previous==')') {
                expr = expr.substring(0, expr.length()-1);
                out.print("<expr"+reverse+">"+expr+"</expr>");
                return;
            }
            if (c=='#' && previous=='#') {
                previous = ' ';
                continue; // ## is translated to #
            }
            previous = c;
            expr += translateChar(c);
        }
        error("expecting end for #(... from line "+lc);
    }
    
    static void grabJstmt(String reverse) {
        int lc = linecount;
        String stmt = "";
        char previous = 'x';
        while (more()) {
            char c = next();
            if (c=='#' /* && (previous==' '||previous=='\n'||previous=='\t') */) {
                //stmt = stmt.substring(0, stmt.length()-1);
                out.print("<java"+reverse+">"+stmt+"</java>");
                return;
            }
            if (c=='#' && previous=='#') {
                previous = 'x';
                continue; // ## is translated to #
            }
            previous = c;
            stmt += translateChar(c);
        }
        error("expecting end for #java on line "+lc);
    }
    
    static void grabFor(String reverse) {
        int lc = linecount;
        if (next()!='r') {
            error("Unknown #for construct");
            return;
        }
        char c = nextnonwhitespace();
        String var = "";
        if (c!='"') {
            var = c+"";
            while (more()) {
                c = next();
                if (c!=' ') {
                    var += c; // shouldn't need translateChar
                } else {
                    break;
                }
            }
            // ==> check for valid var characters
            var = " var=\""+var+"\"";
            c = nextnonwhitespace();
        }
        if (c!='"') {
            error("missing path from #for");
            return;
        }
        String path = getPath();
        if (path==null) {
            error("missing path in #for");
            return;
        }
        out.print("<for"+var+" path="+path+reverse+">");
        if (translate()!="end") {
            error("expecting #end for #for from line "+lc);
            return;
        }
        out.print("</for>");
    }
    
    static void grabIf(String reverse) {
        int lc = linecount;
        char c = nextnonwhitespace();
        if (c!='"') {
            error("missing path in #if");
            return;
        }
        String path = getPath();
        out.print("<if path="+path+reverse+">");
        String n = translate();
        if (n.equals("fi")) {
            out.print("</if>");
        } else if (n.equals("else")) {
            out.print("<else>");
            if (translate().equals("fi")) {
                out.print("</else></if>");
            } else {
                error("expecting #fi# for #if# on line "+lc);
            }
        } else {
            error("expecting #else# or #fi# for #if# on line "+lc);
        }
    }
    
    static void grabDeclare(String reverse) {
        int lc = linecount;
        out.print("<declare"+reverse);
        while (more()) {
            char c = next();
            if (c=='#') {
                out.print("/>");
                return;
            }
            out.print(c); // translateChar ??
        }
        error("terminated in middle of #declare construct on line "+lc);
    }
    
    static void grabDeclarations(String reverse) {
        int lc = linecount;
        out.print("<declarations"+reverse+">");
        while (more()) {
            char c = next();
            if (c=='#') {
                out.print("</declarations>");
                return;
            }
            out.print(translateChar(c));
        }
        error("terminated in middle of #declarations construct on line "+lc);
    }
    
    static String translateChar(char c) {
        switch(c) {
            case '&': return "&amp;";
            case '<': return "&lt;";
            case '>': return "&gt;";
            // case '\'': return "&apos;";
            // case '"': return "&quot;";
            default: return c+"";
        }
    }
        
    
    // Simple set of methods for processing input stream
    
    static boolean more() {
        if (nextchar == -2) {
            try {
                nextchar = in.read();
                if (nextchar == '\n') ++linecount;
            } catch (Exception e) {
                error("Read exception: "+e);
                return false;
            }
        }
        return nextchar != -1;
    }
    
    static char next() {
        if (more()) {
            char c = (char) nextchar;
            nextchar = -2;
            return c;
        }
        return (char)0; // no more
    }
    
    static char nextnonwhitespace() {
        while (more()) {
            char c = next();
            if (c!=' ' && c!='\n' && c!='\t')
                return c;
        }
        return (char)0;
    }
    
    static void error(String m) {
        ++errorcount;
        if (errorcount==20) {
            System.out.println("remaining errors are not printed");
        } else if (errorcount<20) {
            System.out.println("line "+linecount+": "+m);
            out.print("*** Error="+m+" ***");
        }
    }
    
    public static void main(String[] args) {
        try {
            out = args.length>=2
                ?new PrintWriter(new FileOutputStream(args[1]))
                :new PrintWriter(System.out);
            in = args.length>=1
                ? new FileInputStream(args[0]) : System.in;
            out.println("<?xml version=\"1.0\"?>");
            out.print("<tl>");
            translate();
            out.print("</tl>");
            out.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
