/* Copyright (c) 2001, Craig Cleaveland,  All rights reserved.
	http://craigc.com
	
Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the following
  disclaimer.

  Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

  Neither Craig Cleaveland nor the names of its contributors may be
  used to endorse or promote products derived from this software
  without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

import org.w3c.dom.*;
import java.io.*;
import java.util.*;

/**
 * This class translates the XML form of TL to the simple form.
 *
 * The main method is typically invoked on the command line
 * with two arguments.  The first is the name of the input
 * file (XML form) and the second is the name of the output
 * file (simple form).  For example, to translate hello.xml
 * to hello.template, you would say:<pre><code>
 *      java XML2simple hello.xml hello.template
 *</code></pre>
 * @author Craig Cleaveland
 * @version 1.0
 */
public class XML2simple {
    static String delimiter1 = "#";
    static String delimiter2 = "#";
    static boolean genhtml = false;

    boolean translate(XPathContext context, PrintWriter out) {
        try {
            NodeList x = context.evalNodeSet("*|text()");
            for (int j=0; j<x.getLength(); ++j) {
                Node n = (Node) x.item(j);
                context.push(n, j+1, x.getLength());
                switch (n.getNodeType()) {
                case Node.ELEMENT_NODE:
                    String ename = n.getNodeName();
                    String path = translateText("\""+context.evalString("@path")+"\"");
                    String del1 = delimiter1 + (context.evalString("@whitespace").equals("reverse")
                        ? "." : "");
                    if (ename.equals("value")) {
                        out.print(del1+path+delimiter2);
                    } else if (ename.equals("for")) {
                        out.print(del1+"for "+context.evalString("@var")
                            +" "+path+delimiter2);
                        translate(context, out);
                        out.print(delimiter1+"end"+delimiter2);
                    } else if (ename.equals("if")) {
                        out.print(del1+"if "+path+delimiter2);
                        translate(context, out);
                        out.print(delimiter1+"fi"+delimiter2);
                    } else if (ename.equals("else")) {
                        out.print(del1+"else"+delimiter2);
                        translate(context, out);
                    } else if (ename.equals("expr")) {
                        out.print(del1+"(");
                        translate(context, out);
                        out.print(")"+delimiter2);
                    } else if (ename.equals("java")) {
                        out.print(del1+" ");
                        translate(context, out);
                        out.print(delimiter2);
                    } else if (ename.equals("empty")) {
                        out.print(del1+delimiter2);
                    } else if (ename.equals("declarations")) {
                        out.print(del1+"declarations ");
                        translate(context, out);
                        out.print(delimiter2);
                    } else if (ename.equals("declare")) {
                        out.print(del1+"declare");
                        NodeList atts = context.evalNodeSet("@*");
                        for (int k=0; k<atts.getLength(); ++k) {
                            Node att = (Node) atts.item(k);
                            String aname = att.getNodeName();
                            if (!aname.equals("whitespace")) {
                                out.print(" "+aname+"="+translateText(quoteString(att.getNodeValue())));
                            }
                        }
                        out.print(delimiter2);
                    } else {
                        out.print(del1+"UNKNOWN CONSTRUCT: "+ename+delimiter2);
                    }
                    break;
                case Node.TEXT_NODE:
                    out.print(translateText(context.evalString(".")));
                    break;
                }
                context.pop();
            }
        } catch (Exception e) {
            System.out.println("Exception in translate: "+e.getMessage());
            e.printStackTrace();
            return false;
        }
        return true;
    }
    
    String translateText(String s) {
        if (s.indexOf("#")<0
           && (!genhtml || (
              s.indexOf("<")<0
              && s.indexOf("&")<0))) return s;
        // replace all occurrences of # with ##
        String r = "";
        int j;
        while ((j=s.indexOf("#"))>=0) {
            r += s.substring(0, j) + "##";
            s = s.substring(j+1);
        }
        if (!genhtml) return r+s;
        s = r+s;
        r = "";
        while ((j=s.indexOf("&"))>=0) {
            r += s.substring(0, j) + "&amp;";
            s = s.substring(j+1);
        }
        s = r+s;
        r = "";
        while ((j=s.indexOf("<"))>=0) {
            r += s.substring(0, j) + "&lt;";
            s = s.substring(j+1);
        }
        return r+s;
    }
    
    /** add quotes and convert new lines, tabs  quotes */
    String quoteString(String t) {
        String r = "";
        for (int j=0; j<t.length(); ++j) {
            char c = t.charAt(j);
            if (c=='\n') {
                r += "\\n";
            } else if (c=='"') {
                r += "\\\"";
            } else if (c=='\t') {
                r += "\\t";
            } else if (c=='\\') {
                r += "\\\\";
            } else {
                r += c;
            }
        }
        return "\""+r+"\"";
    }
    
    public static void main(String[] args) {
        try {
            String outputFilename = "";
            PrintWriter out = args.length>=2
                ?new PrintWriter(new FileOutputStream(outputFilename = args[1]))
                :new PrintWriter(System.out);
            XML2simple tr = new XML2simple();
            if (outputFilename.endsWith(".html")) {
                delimiter1 = "<font color=red>"+delimiter1;
                delimiter2 += "</font>";
                genhtml = true;
                out.println("<html><head><title>"+args[0]+"</title></head>");
                out.println("<body><h3>"+args[0]+"</h3><pre><code>");
            }
            tr.translate(new XPathContext(args[0]), out);
            if (genhtml) {
                out.println("</code></pre></body></html>");
            }
            out.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
