//import com.jclark.xsl.om.*;
import org.xml.sax.*;
import org.w3c.dom.*;

import java.io.*;
import java.util.*;

import com.fujitsu.xml.omquery.*;

/**
 * This class provides a standard interface to XPath.
 * It is currently implemented using James Clark's xt
 * and Takuki Kamiya's omquery.  An XPathContext is
 * an object representing the information that must be
 * provided for evaluating XPath expressions.  An XPath
 * context includes: current node (sometimes called
 * the context node), two integers (position and size),
 * a set of variables and their values.  It should also
 * include a function library and namesspaces, but these
 * are not provided in this implementation.
 * To facilitate ease of use, XPathContext includes an
 * easy way of saving and restoring the context by using
 * a simple stack.  At this time, only the current node,
 * position and size are saved on the stack.
 *
 * @author Craig Cleaveland
 * @version 1.1
 * 3/30/02 Must explicitly restore variables in setup()
 */
public class XPathContext {
    // An XPath Context includes
    private Document document = null;
    private Node node = null;
    private int size = 1;
    private int position = 1;
    private Hashtable variables = new Hashtable();
    // private function library
    // private namespaces
    
    private DomQueryMgr query_mgr = null;
    private com.fujitsu.xml.omquery.XPathContext ctx = null;
    
    // stacks for saving and restoring context
    private Stack saveStack = new Stack();
    private class SaveContext {
        Node node;
        int position, size;
    }
    
    /**
     * Create a new XPath context.
     * @param d The XML document
     * @param n The current node
     * @param p The position
     * @param s The size
     */
    public XPathContext(Document d, Node n, int p, int s) {
        document = d;
        node = n;
        position = p;
        size = s;
        setup();
    }
    
    void setup() {
        query_mgr = new DomQueryMgr(document);
        ctx = query_mgr.createXPathContext();
        ctx.setSize(size);
        ctx.setPosition(position);
        ctx.setDomNode(node);
        Enumeration e = variables.keys();
        while (e.hasMoreElements()) {
            String key = (String) e.nextElement();
            try { setVariable(key, variables.get(key)); } catch (Exception e2) {}
        }
    }
    
    /**
     * Create XPath context based on another context.
     * Clones a copy of the XPath context.
     * At this time, only the current node, position and size
     * are copied.
     */
    public XPathContext(XPathContext x) {
        document = x.getDocument();
        node = x.getNode();
        position = x.getPosition();
        size = x.getSize();
        setup();
        // what about variables?
    }
    
    /**
     * Create XPath context from an XML source file.
     * The parameters specifies the XML source as a URL.
     * An XML Document will be constructed from the XML file.
     * The current node will be set to the document element
     * and the position and size are set to 1.
     */
    public XPathContext(String xmlSource) throws Exception {
        InputSource input = new InputSource(createURL(xmlSource));
        JAXP_ProcessorImpl xmlproc = new JAXP_ProcessorImpl();
        document = xmlproc.load(input);
        node = document.getDocumentElement();
        position = 1;
        size = 1;
        setup();
    }
    
    /**
     * Evaluates an XPath expression and converts it to a number.
     */
    public double evalNumber(String path) throws Exception {
        return query_mgr.getNumberByXPath(path, ctx);
    }
    
    /**
     * Evaluates an XPath expression and converts it to a boolean.
     */
    public boolean evalBoolean(String path) throws Exception {
        return query_mgr.getBooleanByXPath(path, ctx);
    }
    
    /**
     * Evaluates an XPath expression and converts it to a String.
     */
    public String evalString(String path) throws Exception {
        return query_mgr.getStringByXPath(path, ctx);
    }
    
    /**
     * Evaluates an XPath expression to return a Node Set.
     */
    public NodeList evalNodeSet(String path) throws Exception {
        return new MyNodeList(query_mgr.getNodesByXPath(path, ctx));
    }
    
    /**
     * Push context node, position, and size into the stack
     * and set a new current node, position and size.
     */
    public void push(Node n, int p, int s) {
        SaveContext save = new SaveContext();
        save.node = node;
        save.position = position;
        save.size = size;
        saveStack.push(save);
        node = n;
        position = p;
        size = s;
        setup();
        // should also do something about variable bindings
    }
    
    /**
     * Restore the  context node, position, and size from the stack
     */
    public void pop() throws Exception {
        SaveContext save = (SaveContext) saveStack.pop();
        node = save.node;
        position = save.position;
        size = save.size;
        setup();
        // should also restore variable bindings
    }
        
    /**
     * Set the Document.
     */
    public void setDocument(Document d) {
        document = d;
        setup();
    }
    
    /**
     * Get the Document of the XPath context.
     */
    public Document getDocument() {
        return document;
    }
    
    /**
     * Set the current node of the XPath context.
     */
    public void setNode(Node n) {
        node = n;
        ctx.setDomNode(n);
    }
    
    /**
     * Get the current node of the XPath context.
     */
    public Node getNode() {
        return node;
    }
    
    /**
     * Set the current position of the XPath context.
     */
    public void setPosition(int j) {
        position = j;
        ctx.setPosition(j);
    }
    
    /**
     * Get the current position of the XPath context.
     */
    public int getPosition() {
        return position;
    }
    
    /**
     * Set the current size of the XPath context.
     */
    public void setSize(int j) {
        size = j;
        ctx.setSize(j);
    }
    
    /**
     * Get the current size of the XPath context.
     */
    public int getSize() {
        return size;
    }
    
    /**
     * Set the value of a variable in the XPath context.
     * The value of the variable must be an XPath type which may
     * be any of the following: Double, Node, String, Boolean, or NodeList.
     */
    public void setVariable(String var, Object value) throws Exception {
        variables.put(var, value);
        //com.fujitsu.xml.omquery.XPathContext ctx = query_mgr.createXPathContext();
        if (value instanceof Double) {
            ctx.addNumberVariable(var, ((Double)value).doubleValue());
        } else if (value instanceof Node) {
            ctx.addSingleNodeVariable(var, (Node) value);
        } else if (value instanceof String) {
            ctx.addStringVariable(var, (String) value);
        } else if (value instanceof Boolean) {
            ctx.addBooleanVariable(var, ((Boolean)value).booleanValue());
        } else if (value instanceof NodeList) {
            NodeList nl = (NodeList) value;
            Node[] nodearray = new Node[nl.getLength()];
            for (int j=0; j<nl.getLength(); ++j) {
                nodearray[j] = nl.item(j);
            }
            ctx.addNodeArrayVariable(var, nodearray, 0, nodearray.length);
        } else {
            throw new Exception("setVariable("+var+", "+value.getClass().getName()
                +")");
        }
    }
    
    /**
     * Get the value of a variable in the XPath context.
     * The type of the value will be one of the following:
     * Double, Boolean, String, Node, or NodeList.
     */
    public Object getVariable(String var) {
        return variables.get(var);
    }
        
  private static String createURL(String path)
  {
    File f = new File(path);
    String uri = f.getAbsolutePath();

    char sep = System.getProperty("file.separator").charAt(0);
    uri = uri.replace(sep, '/');
    if (uri.charAt(0) != '/')
      uri = '/' + uri;

    uri = "file://" + uri;

    return uri;
  }

    
}

/**
 * A simple implementation of the DOM type NodeList.
 */
class MyNodeList implements NodeList {
    Vector nodes = new Vector();

    /**
     * Create a NodeList from an enumeration of Nodes
     */
    MyNodeList(Enumeration e) {
        while (e.hasMoreElements()) {
            nodes.addElement(e.nextElement());
        }
    }
    
    /**
     * Get the length of the node list.
     */
    public int getLength() {
        return nodes.size();
    }
    
    /**
     * Get the jth item of the node list (1 indexed).
     */
    public Node item(int j) {
        return (Node) nodes.elementAt(j); // 1-indexed
    }
}
